/* This is free and unencumbered software released into the public domain.
 * 
 * Anyone is free to copy, modify, publish, use, compile, sell, or
 * distribute this software, either in source code form or as a compiled
 * binary, for any purpose, commercial or non-commercial, and by any
 * means.
 *
 * See UNLICENSE for complete information.
 *
 * This FFT-code is part of the implementation which implements 
 * our modified Gauss-sieve algorithm specifically 
 * aimed at finding short vectors in an ideal lattice. Ideal lattices are 
 * lattices with additional structure, namely they are also ideals in a ring R.
 * We restrict to lattices which correspond to ideals in the ring of integers 
 * of a cyclotomic number field, namely R = Z[X]/(X^(2^7)+1). We use this structure
 * by computing 128 scalar products by a single ring product using Nussbaumer's
 * symbolic FFT approach (negacyclic convolution). 
 *
 * See for a detailed description of the methods used:
 * Joppe W. Bos, Michael Naehrig, and Joop van de Pol:
 * Sieving for Shortest Vectors in Ideal Lattices:
 * a Practical Perspective. Cryptology ePrint Archive, 
 * Report 2014/880, IACR, 2014.
 * http://eprint.iacr.org/2014/880.pdf
 */

#include "FFT.h"

DATATYPE ***yy1;
DATATYPE ***z1;
__m128i  **Z1;

static DATATYPE ***x1;
static DATATYPE *t1;
static __m128i  **XX;

/* Reverse the bits, approach from "Bit Twiddling Hacks"
 * See: https://graphics.stanford.edu/~seander/bithacks.html
 */
static uint32_t reverse (uint32_t x)
{
  x = (((x & 0xaaaaaaaa) >> 1) | ((x & 0x55555555) << 1));
  x = (((x & 0xcccccccc) >> 2) | ((x & 0x33333333) << 2));
  x = (((x & 0xf0f0f0f0) >> 4) | ((x & 0x0f0f0f0f) << 4));
  x = (((x & 0xff00ff00) >> 8) | ((x & 0x00ff00ff) << 8));
  return((x >> 16) | (x << 16));
}

/* Nussbaumer approach, see:
 * H. J. Nussbaumer. Fast polynomial transform algorithms for digital 
 * convolution. Acoustics, Speech and Signal Processing, 
 * IEEE Transactions on, 28(2):205{215, 1980.
 * We followed the description from Knuth:
 * D. E. Knuth. Seminumerical Algorithms. 
 * The Art of Computer Programming. Addison-Wesley, Reading,
 * Massachusetts, USA, 3rd edition, 1997
 * Exercise Exercise 4.6.4.59.
 */

void toFFT (DATATYPE ***x2, INT *x) {
  unsigned int i, iter;
  int j;
  
  /* First "level" of FFT */
  for (i=0; i<8; i++) {
    for (j=0; j<16; j++) {
      set(x2[0][i][j],   x[8*j+i]);
      set(x2[0][i+8][j], x[8*j+i]);
    }
  }
 
  for (j=2; j>=0; j--) {
    for (i=0; i<(1U<<(3-j)); i++) {
      unsigned int t;
      for (t=0; t < (1U<<j); t++) {
        unsigned int s, sr, I, L, K, a;
        s = i;
        sr = (reverse (s) >> (29+j));
        sr <<= j;
        s  <<= (j+1);

        // X_i(w) = X_i(w) + w^kX_l(w) can be computed as
        // X_ij = X_ij - X_l(j-k+r)  for  0 <= j < k
        // X_ij = X_ij + X_l(j-k)    for  k <= j < r
        I=s+t, 
        L=s+t+(1<<j); 
        K=2*sr;

        for (a=K; a < 16; a++) set(t1[a], x2[0][L][a-K]);
        for (a=0; a < K; a++)  neg(t1[a], x2[0][L][16+a-K]);

        for (a=0; a<16; a++) {
          sub(x2[0][L][a], x2[0][I][a], t1[a]);
          add(x2[0][I][a], x2[0][I][a], t1[a]);
        }
      }
    }
  }

  for (iter=0; iter < 16; iter++) {
    /* Start the "second" level of forward FFT. */
    for (i=0; i<4; i++) {
      for (j=0; j<4; j++) {
        set(x2[iter+1][i][j],   x2[0][iter][4*j+i]);
        set(x2[iter+1][i+4][j], x2[0][iter][4*j+i]);
      }
    }
  
    for (j=1; j>=0; j--) {
      for (i=0; i<(1U<<(2-j)); i++) {
        unsigned int t;
        for (t=0; t < (1U<<j); t++) {
          unsigned int s, sr, I, L, K, a;
          s=i;
          sr = (reverse (s) >> (32-(2-j)));
          sr <<= j;
          s  <<= (j+1);

          // X_i(w) = X_i(w) + w^kX_l(w) can be computed as
          // X_ij = X_ij - X_l(j-k+r)  for  0 <= j < k
          // X_ij = X_ij + X_l(j-k)    for  k <= j < r
          I=s+t, 
          L=s+t+(1<<j); 
          K=sr;

          for (a=K; a < 4; a++) set(t1[a], x2[iter+1][L][a-K]);
          for (a=0; a < K; a++) neg(t1[a], x2[iter+1][L][4+a-K]);

          for (a=0; a<4; a++) {
            sub(x2[iter+1][L][a], x2[iter+1][I][a], t1[a]);
            add(x2[iter+1][I][a], x2[iter+1][I][a], t1[a]);
          }
        }
      }
    }
  }
}

static void naiveSSE_8 (__m128i *Z, __m128i **x, __m128i **y) {
  __m128i c1, c2, c3, c4, c5, c6, c7, c8, c9, cA, cB, cC;

#define NAIVESSE_LOOP(c1,c2,c3,i) \
    c1 = _mm_madd_epi16 (x[i][0], y[i][0]); \
    c2 = _mm_madd_epi16 (x[i][0], y[i][1]); \
    c3=_mm_unpackhi_epi32(c1,c2); \
    c1=_mm_unpacklo_epi32(c1,c2); \
    c2=_mm_unpackhi_epi32(c1,c3); \
    c1=_mm_unpacklo_epi32(c1,c3); \
    Z[i]=_mm_add_epi32 (c1,c2)    
    
  NAIVESSE_LOOP(c1,c2,c3,0); NAIVESSE_LOOP(c4,c5,c6,1); 
  NAIVESSE_LOOP(c7,c8,c9,2); NAIVESSE_LOOP(cA,cB,cC,3);
  NAIVESSE_LOOP(c1,c2,c3,4); NAIVESSE_LOOP(c4,c5,c6,5); 
  NAIVESSE_LOOP(c7,c8,c9,6); NAIVESSE_LOOP(cA,cB,cC,7);
}
#undef NAIVESSE_LOOP

void FFT_mul(__m128i **Z, __m128i ***x2, __m128i ***y2) {
  unsigned int i;
  for (i=0; i < 16; i++) {
    naiveSSE_8 (Z[i], x2[i], y2[i]);
  }
}

#define LOOP(i) \
  X[i][0] = _mm_load_si128 ((__m128i *) x2[i+1][0]); \
  X[i][1] = _mm_load_si128 ((__m128i *) x2[i+1][1]); \
  X[i][2] = _mm_load_si128 ((__m128i *) x2[i+1][2]); \
  X[i][3] = _mm_load_si128 ((__m128i *) x2[i+1][3]); \
  X[i][4] = _mm_load_si128 ((__m128i *) x2[i+1][4]); \
  X[i][5] = _mm_load_si128 ((__m128i *) x2[i+1][5]); \
  X[i][6] = _mm_load_si128 ((__m128i *) x2[i+1][6]); \
  X[i][7] = _mm_load_si128 ((__m128i *) x2[i+1][7])

#define FFTL2SSE(X,x2) \
  LOOP( 0); LOOP( 1); LOOP( 2); LOOP( 3); \
  LOOP( 4); LOOP( 5); LOOP( 6); LOOP( 7); \
  LOOP( 8); LOOP( 9); LOOP(10); LOOP(11); \
  LOOP(12); LOOP(13); LOOP(14); LOOP(15)

void fromFFT(DATATYPE *x, DATATYPE ***x2, __m128i **X) {
  unsigned int i,iter;
  int j;
  __m128i t0, t2, t3;
  __m128i c0, c1, c2, c3, c4, c5, c6, c7, T2[4];
  
  for (iter=0; iter < 16; iter++) {    
    __m128i T0, T1, T2, T3;  	  
    c1 = _mm_srai_epi32 (_mm_sub_epi32 (X[iter][0], X[iter][1]), 1); 
    c0 = _mm_add_epi32 (c1, X[iter][1]);     
      
    t2 = _mm_srai_epi32 (_mm_sub_epi32 (X[iter][2], X[iter][3]), 1);
    c2 = _mm_add_epi32 (t2, X[iter][3]);
    c3 = _mm_sub_epi32(_mm_srli_si128 (t2, 2*4), _mm_slli_si128 (t2, (4-2)*4));
          
    t3 = _mm_srai_epi32 (_mm_sub_epi32 (X[iter][4], X[iter][5]), 1);
    c4 = _mm_add_epi32 (t3, X[iter][5]);
    c5 = _mm_sub_epi32(_mm_srli_si128 (t3, 1*4), _mm_slli_si128 (t3, (4-1)*4));
          
    t0 = _mm_srai_epi32 (_mm_sub_epi32 (X[iter][6], X[iter][7]), 1);
    c6 = _mm_add_epi32 (t0, X[iter][7]);
    c7 = _mm_sub_epi32(_mm_srli_si128 (t0, 3*4), _mm_slli_si128 (t0, (4-3)*4));
          
    t2 = _mm_srai_epi32 (_mm_sub_epi32 (c0, c2), 1);
    c0 = _mm_add_epi32 (t2, c2);     
    c2 = t2;      
      
    t3 = _mm_srai_epi32 (_mm_sub_epi32 (c1, c3), 1);
    c1 = _mm_add_epi32 (t3, c3);     
    c3 = t3;          
      
    t0 = _mm_srai_epi32 (_mm_sub_epi32 (c4, c6), 1);
    c4 = _mm_add_epi32 (t0, c6);
    c6 = _mm_sub_epi32(_mm_srli_si128 (t0, 2*4), _mm_slli_si128 (t0, (4-2)*4));
          
    t2 = _mm_srai_epi32 (_mm_sub_epi32 (c5, c7), 1);
    c5 = _mm_add_epi32 (t2, c7);
    c7 = _mm_sub_epi32(_mm_srli_si128 (t2, 2*4), _mm_slli_si128 (t2, (4-2)*4));
          
    t3 = _mm_srai_epi32 (_mm_sub_epi32 (c0, c4), 1);
    c0 = _mm_add_epi32 (t3, c4);     
    c4 = t3;
      
    t0 = _mm_srai_epi32 (_mm_sub_epi32 (c1, c5), 1);
    c1 = _mm_add_epi32 (t0, c5);     
    c5 = t0;
      
    t2 = _mm_srai_epi32 (_mm_sub_epi32 (c2, c6), 1);
    c2 = _mm_add_epi32 (t2, c6);     
    c6 = t2;
      
    t3 = _mm_srai_epi32 (_mm_sub_epi32 (c3, c7), 1);
    c3 = _mm_add_epi32 (t3, c7);     
    c7 = t3;
      
    T0 = _mm_unpacklo_epi32(c0, c1);
    T1 = _mm_unpacklo_epi32(c2, c3);
    T2 = _mm_unpackhi_epi32(c0, c1);
    T3 = _mm_unpackhi_epi32(c2, c3);

    /* Assigning transposed values back */
    c0 = _mm_unpacklo_epi64(T0, T1);
    c1 = _mm_unpackhi_epi64(T0, T1);
    c2 = _mm_unpacklo_epi64(T2, T3);
    c3 = _mm_unpackhi_epi64(T2, T3);
    
    T0 = _mm_unpacklo_epi32(c4, c5);
    T1 = _mm_unpacklo_epi32(c6, c7);
    T2 = _mm_unpackhi_epi32(c4, c5);
    T3 = _mm_unpackhi_epi32(c6, c7);
    
    c4 = _mm_unpacklo_epi64(T0, T1);
    c5 = _mm_unpackhi_epi64(T0, T1);
    c6 = _mm_unpacklo_epi64(T2, T3);
    c7 = _mm_unpackhi_epi64(T2, T3);
    
    XX[iter][0] = _mm_sub_epi32(c0,c7);
    XX[iter][1] = _mm_add_epi32(c1,c4);
    XX[iter][2] = _mm_add_epi32(c2,c5);
    XX[iter][3] = _mm_add_epi32(c3,c6);    
  }

#define LOOP1(A,B,a)                     \
  t1[a] = (x2[0][A][a] - x2[0][B][a])/2; \
  x2[0][A][a] = t1[a] + x2[0][B][a]

#define LOOP2(A,B,i)  LOOP1(A,B,i); LOOP1(A,B,i+1);
#define LOOP4(A,B,i)  LOOP2(A,B,i); LOOP2(A,B,i+2);
#define LOOP8(A,B,i)  LOOP4(A,B,i); LOOP4(A,B,i+4);
#define LOOP16(A,B,i) LOOP8(A,B,i); LOOP8(A,B,i+8);
  
#define UNROLL(A,B,K)                                \
  LOOP16(A,B,0);                                     \
  for (a=0; a<16-K; a++) x2[0][B][a] = t1[a+K];      \
  for(a=16-K; a<16; a++) x2[0][B][a] = -t1[a-(16-K)]
    
#define LOOP16_SSE(A,B)                                           \
  T2[0] = _mm_srai_epi32 (_mm_sub_epi32 (XX[A][0], XX[B][0]), 1); \
  T2[1] = _mm_srai_epi32 (_mm_sub_epi32 (XX[A][1], XX[B][1]), 1); \
  T2[2] = _mm_srai_epi32 (_mm_sub_epi32 (XX[A][2], XX[B][2]), 1); \
  T2[3] = _mm_srai_epi32 (_mm_sub_epi32 (XX[A][3], XX[B][3]), 1); \
  XX[A][0] = _mm_add_epi32 (T2[0], XX[B][0]);                     \
  XX[A][1] = _mm_add_epi32 (T2[1], XX[B][1]);                     \
  XX[A][2] = _mm_add_epi32 (T2[2], XX[B][2]);                     \
  XX[A][3] = _mm_add_epi32 (T2[3], XX[B][3])
  
#define UNROLL_SSE_Ke0(A,B,K) \
  LOOP16_SSE(A,B);            \
  XX[B][0] = T2[0];           \
  XX[B][1] = T2[1];           \
  XX[B][2] = T2[2];           \
  XX[B][3] = T2[3]

#define UNROLL_SSE_Ke8(A,B,K)                           \
  LOOP16_SSE(A,B);                                      \
  XX[B][0] = T2[2];                                     \
  XX[B][1] = T2[3];                                     \
  XX[B][2] = _mm_sub_epi32(_mm_setzero_si128(), T2[0]); \
  XX[B][3] = _mm_sub_epi32(_mm_setzero_si128(), T2[1])

#define UNROLL_SSE_Ke4(A,B,K)                           \
  LOOP16_SSE(A,B);                                      \
  XX[B][0] = T2[1];                                     \
  XX[B][1] = T2[2];                                     \
  XX[B][2] = T2[3];                                     \
  XX[B][3] = _mm_sub_epi32(_mm_setzero_si128(), T2[0])
  
#define UNROLL_SSE_Ke12(A,B,K)                          \
  LOOP16_SSE(A,B);                                      \
  XX[B][0] = T2[3];                                     \
  XX[B][1] = _mm_sub_epi32(_mm_setzero_si128(), T2[0]); \
  XX[B][2] = _mm_sub_epi32(_mm_setzero_si128(), T2[1]); \
  XX[B][3] = _mm_sub_epi32(_mm_setzero_si128(), T2[2])

#define UNROLL_SSE_K1t3(A,B,K)                                                             \
  LOOP16_SSE(A,B);                                                                         \
  XX[B][0] = _mm_add_epi32( _mm_srli_si128 (T2[0], K*4), _mm_slli_si128 (T2[1], (4-K)*4)); \
  XX[B][1] = _mm_add_epi32( _mm_srli_si128 (T2[1], K*4), _mm_slli_si128 (T2[2], (4-K)*4)); \
  XX[B][2] = _mm_add_epi32( _mm_srli_si128 (T2[2], K*4), _mm_slli_si128 (T2[3], (4-K)*4)); \
  XX[B][3] = _mm_sub_epi32( _mm_srli_si128 (T2[3], K*4), _mm_slli_si128 (T2[0], (4-K)*4))
    
#define UNROLL_SSE_K9t11(A,B,K)                                                                                                       \
  LOOP16_SSE(A,B);                                                                                                                    \
  XX[B][0] = _mm_add_epi32( _mm_srli_si128 (T2[2], (K-8)*4), _mm_slli_si128 (T2[3], (12-K)*4));                                       \
  XX[B][1] = _mm_sub_epi32( _mm_srli_si128 (T2[3], (K-8)*4), _mm_slli_si128 (T2[0], (12-K)*4));                                       \
  XX[B][2] = _mm_sub_epi32( _mm_setzero_si128(), _mm_add_epi32( _mm_srli_si128 (T2[0], (K-8)*4), _mm_slli_si128 (T2[1], (12-K)*4)));  \
  XX[B][3] = _mm_sub_epi32( _mm_setzero_si128(), _mm_add_epi32( _mm_srli_si128 (T2[1], (K-8)*4), _mm_slli_si128 (T2[2], (12-K)*4)))
                
#define UNROLL_SSE_K5t7(A,B,K)                                                                                                        \
  LOOP16_SSE(A,B);                                                                                                                    \
  XX[B][0] = _mm_add_epi32( _mm_srli_si128 (T2[1], (K-4)*4), _mm_slli_si128 (T2[2], (8-K)*4));                                        \
  XX[B][1] = _mm_add_epi32( _mm_srli_si128 (T2[2], (K-4)*4), _mm_slli_si128 (T2[3], (8-K)*4));                                        \
  XX[B][2] = _mm_sub_epi32(_mm_srli_si128 (T2[3], (K-4)*4), _mm_slli_si128 (T2[0], (8-K)*4));                                         \
  XX[B][3] = _mm_sub_epi32(_mm_setzero_si128(), _mm_add_epi32( _mm_srli_si128 (T2[0], (K-4)*4), _mm_slli_si128 (T2[1], (8-K)*4)))
 
#define UNROLL_SSE_K13t15(A,B,K)                                                                                                      \
  LOOP16_SSE(A,B);                                                                                                                    \
  XX[B][0] = _mm_sub_epi32( _mm_srli_si128 (T2[3], (K-12)*4), _mm_slli_si128 (T2[0], (16-K)*4));                                      \
  XX[B][1] = _mm_sub_epi32( _mm_setzero_si128(), _mm_add_epi32( _mm_srli_si128 (T2[0], (K-12)*4), _mm_slli_si128 (T2[1], (16-K)*4))); \
  XX[B][2] = _mm_sub_epi32( _mm_setzero_si128(), _mm_add_epi32( _mm_srli_si128 (T2[1], (K-12)*4), _mm_slli_si128 (T2[2], (16-K)*4))); \
  XX[B][3] = _mm_sub_epi32( _mm_setzero_si128(), _mm_add_epi32( _mm_srli_si128 (T2[2], (K-12)*4), _mm_slli_si128 (T2[3], (16-K)*4)))
        
#define UNROLL_SSE_Kg8(A,B,K)                                                                                                         \
  LOOP16_SSE(A,B);                                                                                                                    \
  XX[B][0] = _mm_sub_epi16(_mm_srli_si128 (T2[1], (K-8)*2), _mm_slli_si128 (T2[0], (K)*2));                                           \
  XX[B][1] = _mm_sub_epi16(_mm_setzero_si128(), _mm_add_epi16( _mm_srli_si128 (T2[0], (K-8)*2), _mm_slli_si128 (T2[1], (K)*2)))

  UNROLL_SSE_Ke0   ( 0,  1,  0);
  UNROLL_SSE_Ke8   ( 2,  3,  8);
  UNROLL_SSE_Ke4   ( 4,  5,  4);
  UNROLL_SSE_Ke12  ( 6,  7, 12);
  UNROLL_SSE_K1t3  ( 8,  9,  2);    
  UNROLL_SSE_K9t11 (10, 11, 10);
  UNROLL_SSE_K5t7  (12, 13,  6);
  UNROLL_SSE_K13t15(14, 15, 14);
  UNROLL_SSE_Ke0   ( 0,  2,  0);  
  UNROLL_SSE_Ke0   ( 1,  3,  0);  
  UNROLL_SSE_Ke8   ( 4,  6,  8);
  UNROLL_SSE_Ke8   ( 5,  7,  8);
  UNROLL_SSE_Ke4   ( 8, 10,  4);
  UNROLL_SSE_Ke4   ( 9, 11,  4);
  UNROLL_SSE_Ke12  (12, 14, 12);
  UNROLL_SSE_Ke12  (13, 15, 12);
  UNROLL_SSE_Ke0   ( 0,  4,  0);
  UNROLL_SSE_Ke0   ( 1,  5,  0);
  UNROLL_SSE_Ke0   ( 2,  6,  0);
  UNROLL_SSE_Ke0   ( 3,  7,  0);
  UNROLL_SSE_Ke8   ( 8, 12,  8);
  UNROLL_SSE_Ke8   ( 9, 13,  8);
  UNROLL_SSE_Ke8   (10, 14,  8);
  UNROLL_SSE_Ke8   (11, 15,  8);
  UNROLL_SSE_Ke0   ( 0,  8,  0);
  UNROLL_SSE_Ke0   ( 1,  9,  0);
  UNROLL_SSE_Ke0   ( 2, 10,  0);
  UNROLL_SSE_Ke0   ( 3, 11,  0);
  UNROLL_SSE_Ke0   ( 4, 12,  0);
  UNROLL_SSE_Ke0   ( 5, 13,  0);
  UNROLL_SSE_Ke0   ( 6, 14,  0);
  UNROLL_SSE_Ke0   ( 7, 15,  0);
  
  for (i=0; i < 16; i++) {
    _mm_store_si128 ((__m128i *)  x2[0][i],     XX[i][0]);
    _mm_store_si128 ((__m128i *) &x2[0][i][4],  XX[i][1]);
    _mm_store_si128 ((__m128i *) &x2[0][i][8],  XX[i][2]);
    _mm_store_si128 ((__m128i *) &x2[0][i][12], XX[i][3]);
  }

  for (i=0; i < 8; i++) {
    sub(x[i], x2[0][i][0], x2[0][8+i][15]);
    for (j=1; j < 16; j++) {  
      add(x[8*j+i], x2[0][i][j], x2[0][8+i][j-1]);
    }
  }

#undef LOOP1
#undef LOOP4
#undef LOOP8
#undef LOOP16
#undef UNROLL
}

void FFT_init () {
  int i,j;      
  
  posix_memalign ((void *) &x1,  16, 17 * sizeof (DATATYPE **));
  posix_memalign ((void *) &yy1, 16, 17 * sizeof (DATATYPE **));
  posix_memalign ((void *) &z1,  16, 17 * sizeof (DATATYPE **));
  posix_memalign ((void *) &Z1,  16, 17 * sizeof (__m128i *));
  posix_memalign ((void *) &XX,  16, 16 * sizeof (__m128i *));
  
  posix_memalign ((void *) &x1[0],  16, 16 * sizeof (DATATYPE *));
  posix_memalign ((void *) &yy1[0], 16, 16 * sizeof (DATATYPE *));
  posix_memalign ((void *) &z1[0],  16, 16 * sizeof (DATATYPE *));
  posix_memalign ((void *) &Z1[0],  16, 16 * sizeof (__m128i));
  
  for (i=0; i < 16; i++) {
    posix_memalign ((void *) &x1[0][i],  16, 16 * sizeof (DATATYPE));
    posix_memalign ((void *) &yy1[0][i], 16, 16 * sizeof (DATATYPE));
    posix_memalign ((void *) &z1[0][i],  16, 16 * sizeof (DATATYPE));
    posix_memalign ((void *) &XX[i],     16,  4 * sizeof (__m128i));
  }
  
  for(i = 1; i < 17; i++){
    posix_memalign ((void *) &x1[i],  16, 8 * sizeof (DATATYPE *));
    posix_memalign ((void *) &yy1[i], 16, 8 * sizeof (DATATYPE *));
    posix_memalign ((void *) &z1[i],  16, 8 * sizeof (DATATYPE *));
    posix_memalign ((void *) &Z1[i],  16, 8 * sizeof (__m128i));
    for (j=0; j < 8; j++) {
      posix_memalign ((void *) &x1[i][j],  16, 4 * sizeof (DATATYPE));
      posix_memalign ((void *) &yy1[i][j], 16, 4 * sizeof (DATATYPE));
      posix_memalign ((void *) &z1[i][j],  16, 4 * sizeof (DATATYPE));
    }
    posix_memalign ((void *) &t1, 16, 16 * sizeof (DATATYPE));
  }
}

void FFT_free () {
  int i,j;

  free (t1);
  for (i=0; i < 16; i++) {
    free (x1[0][i]);
    free (yy1[0][i]);
    free (z1[0][i]);
  }
  free (x1[0]);
  free (yy1[0]);
  free (z1[0]);
  
  for (i=1; i < 17; i++){
    for (j=0; j < 8; j++) {
      free (x1[i][j]);
      free (yy1[i][j]);
      free (z1[i][j]);
    }
    free (x1[i]);
    free (yy1[i]);
    free (z1[i]);
  }

  free (x1);
  free (yy1);
  free (z1);
}

#define LOAD1(c,a) \
  c[0] = a[0];     \
  c[1] = a[1];     \
  c[2] = a[2];     \
  c[3] = a[3];     \
  c[4] = a[0];     \
  c[5] = a[1];     \
  c[6] = a[2];     \
  c[7] = a[3]

#define LOAD2(c,a) \
  c[0] = a[0];     \
  c[1] = -a[3];    \
  c[2] = -a[2];    \
  c[3] = -a[1];    \
  c[4] = a[1];     \
  c[5] = a[0];     \
  c[6] = -a[3];    \
  c[7] = -a[2]

#define LOAD3(c,a) \
  c[0] = a[2];     \
  c[1] = a[1];     \
  c[2] = a[0];     \
  c[3] = -a[3];    \
  c[4] = a[3];     \
  c[5] = a[2];     \
  c[6] = a[1];     \
  c[7] = a[0]

/* Put the newvFFT in SSE-form */
void FFT2SSE1(__m128i ***fft, DATATYPE ***x2) {
  int16_t T1[8];
  int i,j;
  for(i=0; i<16; i++) {
    for(j=0;j<8;j++) {
      LOAD1(T1,x2[i+1][j]);
      fft[i][j][0] = _mm_loadu_si128 ((__m128i *) T1);
    }
  }
}

/* Put the listFFT in SSE-form */
void FFT2SSE2(__m128i ***fft, DATATYPE ***x2) {
  int16_t T2[8], T3[8];
  int i,j;
  for(i=0; i<16; i++) {
    for(j=0;j<8;j++) {
      LOAD2(T2,x2[i+1][j]);
      fft[i][j][0] = _mm_loadu_si128 ((__m128i *) T2);
      LOAD3(T3,x2[i+1][j]);  
      fft[i][j][1] = _mm_loadu_si128 ((__m128i *) T3);
    }
  }
}

