/* This is free and unencumbered software released into the public domain.
 * 
 * Anyone is free to copy, modify, publish, use, compile, sell, or
 * distribute this software, either in source code form or as a compiled
 * binary, for any purpose, commercial or non-commercial, and by any
 * means.
 *
 * See UNLICENSE for complete information.
 *
 * This code implements our modified Gauss-sieve algorithm specifically 
 * aimed at finding short vectors in an ideal lattice. Ideal lattices are 
 * lattices with additional structure, namely they are also ideals in a ring R.
 * We restrict to lattices which correspond to ideals in the ring of integers 
 * of a cyclotomic number field, namely R = Z[X]/(X^(2^7)+1). We use this structure
 * by computing 128 scalar products by a single ring product using Nussbaumer's
 * symbolic FFT approach (negacyclic convolution). 
 *
 * See for a detailed description of the methods used:
 * Joppe W. Bos, Michael Naehrig, and Joop van de Pol:
 * Sieving for Shortest Vectors in Ideal Lattices:
 * a Practical Perspective. Cryptology ePrint Archive, 
 * Report 2014/880, IACR, 2014.
 * http://eprint.iacr.org/2014/880.pdf
 */

#include "FFT.h"

/* Saves the list every STORE rounds or never if 0 */
#define STORE 0

/* Size increments of main list and stack respectively */
#define SIZE (8*1024)
#define SIZE2 1024

/* Starting list size */
#define INIT_SIZE (128)

/* Samples to be used per round per node */
#define BATCH (16)

extern DATATYPE ***yy1;
extern DATATYPE ***z1;
extern __m128i  **Z1;

static FILE *rout;

typedef struct {
  INT *v;
  IP inprod;
  __m128i ***fft;
} element_t;

static void element_init (element_t *new) {
  unsigned int i,j;
  new->v = (INT *) malloc ((N) * sizeof (INT));
  if (new->v == NULL) {
    fprintf (stderr, "Error: element_init out of memory.\n");
    exit (-1);
  }
  new->fft = (__m128i ***) malloc (16 * sizeof (__m128i **));
  if (new->fft == NULL) {
    fprintf (stderr, "Error: element_init out of memory.\n");
    exit (-1);
  }
  for(i=0;i<16;i++){
    new->fft[i] = (__m128i **) malloc (8 * sizeof (__m128i *));
    if (new->fft[i] == NULL) {
      fprintf (stderr, "Error: element_init out of memory.\n");
      exit (-1);
    }
    for(j=0;j<8;j++){
      new->fft[i][j] = (__m128i *) malloc (2 * sizeof (__m128i));
      if (new->fft[i][j] == NULL) {
        fprintf (stderr, "Error: element_init out of memory.\n");
        exit (-1);
      }
    }
  }
}

static void copy (INT *a, INT *b, int n) {
  memcpy(a,b,n*sizeof(INT));
}

static void copy_element (element_t *c, element_t *a) {
  memcpy (c->v, a->v, N*sizeof(INT));
  c->inprod = a->inprod;
}

static IP innerproduct (INT *a, INT *b) {
  int i;
  IP sum = a[0]*b[0];
  for(i=1;i<N;i++){
    sum+=a[i]*b[i];
  }
  return sum;
}

/* a is rotated left */
static void vec_submul(INT *a, IP c, INT *b, int index){
  int i;
  
  for(i=0; i < N-index; i++){
    a[i+index]-=c*b[i];
  }
  for(; i < N; i++){
    a[i+index-N]+=c*b[i];
  }
}

/* b is rotated left */
static void vec_submul2(INT *a, IP c, INT *b, int index){
  int i;
  
  for(i=0; i < N-index; i++){
    a[i]-=c*b[i+index];
  }
  for(; i < N; i++){
    a[i]+=c*b[i+index-N];
  }
}

static int reduce(element_t *list, element_t *samp){
  IP inprodnew, inprodnewa, absip; 
  int index,i,order;
  static DATATYPE *z;
  static int init = 1;

  if (init) {
    init = 0;
    z = (DATATYPE *) malloc (N * sizeof(DATATYPE));
  }

  order = (samp->inprod < list->inprod);
  
  inprodnewa = 0;
  inprodnew = 0; 
  FFT_mul(Z1, samp->fft, list->fft);
  fromFFT(z,z1,Z1);
  index = 0;
  for(i=0;i<N;i++){      
    absip = abs(z[i]);
    absip <<=1;
    if(absip > inprodnewa){
      inprodnewa = absip;
      index = N-1-i;
      inprodnew = z[i];
    }
  }
  
  if (order && inprodnewa > samp->inprod) { 
    INT mu; 
    mu = ((inprodnew^list->inprod)<0); mu = -mu*2+1; 
    vec_submul2 (list->v, mu, samp->v, index); 
    UPDATE_MU_STAR(list,samp)
    //if(list->inprod != innerproduct(list->v,list->v)) printf("Error2!\n");
    
    //Vector list has been reduced
    return 1;
  } else if (!order && inprodnewa > list->inprod) { 
    INT mu; 
    mu = ((inprodnew^list->inprod)<0); mu = -mu*2+1; 
    vec_submul (samp->v, mu, list->v, index); 
    UPDATE_MU_STAR(samp,list) 
    //if(samp->inprod != innerproduct(samp->v,samp->v)) printf("Error!\n");
    
    //Vector samp has been reduced
    return 2;
  }
  
  return 0;
}

#if STORE
static void store_element (element_t *element, FILE *out) {
  fwrite (element->v, sizeof (__m128i), (N/8), out);
  fwrite (&element->inprod, sizeof(IP), 1, out);
}

static int store_state (char *fname, int stacksize, int Ssize, element_t *S, \
                 int listsize, int Lsize, int oldLsize, element_t *L, \
                 int64_t *L_bits, int64_t minimum, int iterations, int collisions) {
  char str[1024];
  FILE *out;
  int i;

  /* Keep always an old and a new copy around just to be sure. */
  sprintf (str, "(mv %s %s.old 2>&1) >> /dev/null", fname, fname);
  system (str);

  out = fopen (fname, "w");
  if (out == NULL) return -1;

  /* Write the stack. */
  fwrite (&stacksize, sizeof(int), 1, out);
  fwrite (&Ssize, sizeof(int), 1, out);
  for (i=0; i < stacksize; i++) {
    store_element (&S[i], out);
  }

  /* Write the list. */
  fwrite (&listsize, sizeof(int), 1, out);
  fwrite (&Lsize, sizeof(int), 1, out);
  fwrite (&oldLsize, sizeof(int), 1, out);
  for (i=0; i < listsize; i++) {
    store_element (&L[i], out);
    //printf ("Store: %ld\n", L[i].inprod);
  }

  fwrite (L_bits, sizeof(int64_t), (Lsize/64+1), out);

  fwrite (&minimum, sizeof(int64_t), 1, out);
  fwrite (&iterations, sizeof(int), 1, out);
  fwrite (&collisions, sizeof(int), 1, out);

  fclose (out);
  return 0;
}
#endif

static void load_element (element_t *element, FILE *out) {
  fread (element->v, sizeof (__m128i), (N/8), out);
  fread (&element->inprod, sizeof(IP), 1, out);
  if (innerproduct(element->v,element->v) != element->inprod) {
    printf ("Read error\n");
    exit (0);
  }
}

static int load_state (char *fname, int *stacksize, int *Ssize, element_t **S,  \
                       int *listsize, int *Lsize, int *oldLsize, element_t **L, \
                       int64_t **L_bits, int64_t *minimum, int *iterations, int *collisions) {
  FILE *out;
  int i;

  out = fopen (fname, "r");
  if (out == NULL) return -1;

  /* Read the stack. */
  fread (stacksize, sizeof(int), 1, out);
  fread (Ssize, sizeof(int), 1, out);
  (*S) = (element_t *) malloc ((*Ssize) * sizeof (element_t));
  for(i=0; i<(*Ssize); i++) element_init (&(*S)[i]);
  for (i=0; i<(*stacksize); i++) {
    load_element (&(*S)[i], out);
  }

  /* Read the list. */
  fread (listsize, sizeof(int), 1, out);
  fread (Lsize, sizeof(int), 1, out);
  fread (oldLsize, sizeof(int), 1, out);
  (*L) = (element_t *) malloc ((*Lsize) * sizeof (element_t));
  for(i=0; i<(*Lsize); i++) element_init (&(*L)[i]);
  for (i=0; i<(*listsize); i++) {
    load_element (&(*L)[i], out);
    toFFT(yy1,(*L)[i].v);
    FFT2SSE2((*L)[i].fft,yy1);  
  }

  *L_bits = (int64_t *) malloc (((*Lsize)/64+1)*sizeof(int64_t));
  fread (*L_bits, sizeof(int64_t), ((*Lsize)/64+1), out);

  fread (minimum, sizeof(int64_t), 1, out);
  fread (iterations, sizeof(int), 1, out);
  fread (collisions, sizeof(int), 1, out);

  fclose (out);
  return 0;
}

int client(int rank, int size){
  MPI_Status status;
  INT **b = NULL; //Basis, used for sampling
  INT *sampv = NULL; //Vector, used for sampling
  element_t *L = NULL; //List
  int64_t *L_bits = NULL; //Mask indicating list occupancy
  int listsize; //Number of vectors currently in the list
  int Lsize, oldLsize; //Current list storage size
  
  element_t *S = NULL; //Stack
  int stacksize; //Number of vectors currently in the stack
  int Ssize; //Current stack storage size
  element_t new;
  
  element_t *Q = NULL; //Sample list
  int Qsize = BATCH*size;
  int Qoffset = BATCH*rank;
  int QoffsetB = Qoffset+BATCH;
  
  int16_t *com_buffer = NULL; //Used to send vectors
  int comSize;
  int64_t *survivors = NULL, *rsurv = NULL; //Masks indicating survivors
  int64_t *reqsurv = NULL, *reqrecv = NULL; //Masks for requesting vectors
  int32_t *reqnums = NULL;
  
  int i, j, k, l;
  int64_t max, minimum, goal;
  double **gs, **mu, *lbj = NULL, *vars = NULL, t; //Gram-Schmidt used for sampling
  int collisions, iterations;
  int terminate;
  int16_t svp[N];
  double starttime, inittime, mintime, ghtime;
  
  struct { 
    int32_t lsize; 
    int32_t rank; 
  } lrsend, lrrecv;

  struct _inprod_rank {
    int64_t inprod;
    int32_t rank;
  } *irsend = NULL, *irrecv = NULL, minsend, minrecv;
 
  char fname[1024];
  int restore = 0;
  INT x[N];

  irsend = (struct _inprod_rank*) malloc (Qsize * sizeof(struct _inprod_rank));
  irrecv = (struct _inprod_rank*) malloc (Qsize * sizeof(struct _inprod_rank));
  reqnums = (int32_t *) malloc (size * sizeof(int32_t));
  if (irsend == NULL || irrecv == NULL || reqnums == NULL) {
    fprintf (stderr, "Memory error\n");
    return -1;
  }

  sprintf (fname, "FFTSSE_store.rank.%d", rank);

  //Randomize seed in a naive way
  { 
    struct timeval time; 
    gettimeofday(&time,NULL);
    srand( ((rank << 16) * getpid()) ^ ((time.tv_sec * 1000) + (time.tv_usec / 1000)) ); 
  } 

  /* Try to open a stored state. */
  {
    FILE *try = fopen (fname, "r");
    if (try != NULL) {
      restore = 1;
      fclose (try);
    }
  }

  element_init(&new);
  for(i=0; i < N; i++) new.v[i] = 0;
  new.inprod = 0;
  
  //Initialize memory
  FFT_init();
  starttime = MPI_Wtime();
  if (restore) {
    load_state (fname, &stacksize, &Ssize, &S, \
                &listsize, &Lsize, &oldLsize, &L, &L_bits, &minimum, &iterations, &collisions);
  } else {
    Lsize = SIZE;
    oldLsize = 0;
    L = (element_t *) malloc (SIZE * sizeof (element_t));
    for(i=0; i < SIZE; i++){
      element_init((L+i));
    }

    L_bits = (int64_t *) malloc ((SIZE/64+1)*sizeof(int64_t));
  
    Ssize = SIZE2;
    S = (element_t *) malloc (SIZE2 * sizeof (element_t));
    for(i=0; i < SIZE2; i++){
      element_init((S+i));
    }
  }

#define INCREMENT_LISTSIZE                                             \
listsize++;                                                            \
if(listsize == Lsize) {                                                \
  oldLsize = Lsize;                                                    \
  Lsize += SIZE;                                                       \
  L = (element_t *) realloc (L, Lsize*sizeof(element_t));              \
  for(k=oldLsize; k < Lsize; k++) element_init(L+k);                   \
  L_bits = (int64_t *) realloc (L_bits, (Lsize/64+1)*sizeof(int64_t)); \
  for (l=oldLsize; l < Lsize; l++) { clearbit(L_bits,l); }             \
}

#define INCREMENT_STACKSIZE                                    \
stacksize++;                                                   \
if(stacksize == Ssize) {                                       \
  l = Ssize;                                                   \
  Ssize += SIZE2;                                              \
  S = (element_t *) realloc (S, (Ssize) * sizeof (element_t)); \
  for(;l<Ssize; l++) element_init((S+l));                      \
}

  Q = (element_t *) malloc (Qsize * sizeof (element_t));
  for(i=0; i < Qsize; i++){
    element_init((Q+i));
  }
  
  comSize = N*Qsize;
  com_buffer = (int16_t *) malloc ((N*Qsize) * sizeof (int16_t));
  
  survivors = (int64_t *) malloc ((Qsize/64+1)*sizeof(int64_t));
  rsurv = (int64_t *) malloc ((Qsize/64+1)*sizeof(int64_t));
  reqsurv = (int64_t *) malloc((Qsize/64+1) * size * sizeof(int64_t));
  reqrecv = (int64_t *) malloc((Qsize/64+1) * sizeof(int64_t));
  
  for(i = 0; i < Qsize; i++) irsend[i].rank = rank;
  
  // Initialize memory for basis
  b = (INT**) malloc (n * sizeof (INT*));
  for (i=0; i < n; i++) {
    b[i]  = (INT*) malloc (N * sizeof (INT));
  }
  posix_memalign ((void *) &(sampv), 16, (N) * sizeof (INT));
  
  max = readbasis (fileName, b);
  if (max == 0) {
    printf ("Read basis error\n");
    exit (0);
    return -1;
  }
  
  gs  = (double **) malloc (N * sizeof (double*));
  mu  = (double **) malloc (N * sizeof (double*));
  lbj = (double *)  malloc (N * sizeof (double));
  vars = (double *)  malloc (N * sizeof (double));
        
  for (i=0; i < N; i++) {
    gs[i] = (double *)  malloc (N * sizeof (double));
    mu[i] = (double *)  malloc (N * sizeof (double));
  }

  gram_schmidt (gs, mu, lbj, b, &goal);
  sample_init(lbj,vars);

  t=(log((double)N)*log((double)N))/70.; // ToDO: find the optimal value for t  
  
  terminate = 0;
  
  inittime = 0;
  mintime = 0;
  ghtime = -1.0;
  for(i=0; i<N; i++) svp[i] = 0;
  
  if (!restore) {
    copy(new.v,b[0],N);
    minimum = innerproduct(new.v,new.v);
  
    //Populate the list with the first element 
    if(rank < N){
      copy(L[0].v,b[rank],N);
    } else {
      sample (L[0].v,b,mu,vars,t);
    }
    setbit(L_bits,0);
    L[0].inprod = innerproduct(L[0].v,L[0].v); 
    toFFT(yy1,L[0].v);
    FFT2SSE2(L[0].fft,yy1);
    listsize = 1;
  
    if(rank == 0) fprintf(rout, "Start minimum %lu\n", minimum);
    //Now fill the stack up to INIT_SIZE-1
    stacksize = 0;
    for(i=rank+size; i < N; i+=size){
      copy(S[stacksize].v,b[i],N);
      S[stacksize].inprod = innerproduct(S[stacksize].v,S[stacksize].v);
      stacksize++;
    }
  
    for(;stacksize < INIT_SIZE-1; stacksize++){
      sample (S[stacksize].v,b,mu,vars,t);
      S[stacksize].inprod = innerproduct(S[stacksize].v,S[stacksize].v);
    }
  
    k=0;
  
    //Perform reduction until stack is empty
    collisions = 0;
    while(stacksize > 0){ 
      int firstempty = Lsize;
      j=0;
      copy_element(&new, (S+stacksize-1));
      stacksize--;
      for(i=0; i<N; i++) x[N-1-i] = new.v[i];
      toFFT(yy1,x);
      FFT2SSE1(new.fft,yy1);
      for(i = 0; i < Lsize; i++){
        if(getbit(L_bits, i) != 0){
          int red = reduce((L+i),&new);
          j++;
          if (red & 1) { 
            clearbit (L_bits, i);
            listsize--; j--; 
            if (L[i].inprod == 0) { collisions++; } 
            else { 
              copy_element((S+stacksize),(L+i));
              INCREMENT_STACKSIZE
            }
          } else if (red & 2) {
            if (new.inprod != 0) {
              for(l=0; l<N; l++) { x[N-1-l] = new.v[l]; }
              toFFT(yy1,x);
              FFT2SSE1(new.fft,yy1);
              i=-1; j=0; continue;
            } else { collisions++; break; }
          }
        } else if(i < firstempty) firstempty = i;
      
        if(j == listsize && (firstempty < Lsize)) break;
      }
      if(new.inprod == 0){
        continue;
      } else if(new.inprod < minimum){
        minimum = new.inprod;
        printf("Rank %d: New minimum %lu\n", rank, minimum);
        if(minimum < goal) {
          if(ghtime < 0) ghtime = MPI_Wtime()-starttime;
           printf("Goal reached! Entering termination.\n");
          copy(svp,L[i].v,N);
          terminate = 1;
        }
      }
      
      if (firstempty == Lsize){
        //Error
        fprintf (stderr, "Error: list error.\n");
        exit(1);
      }
      copy_element((L+firstempty), &new);
      setbit(L_bits,firstempty);
      toFFT(yy1,L[firstempty].v);
      FFT2SSE2(L[firstempty].fft,yy1);  
      INCREMENT_LISTSIZE
    }
    
  }
 
  MPI_Barrier (MPI_COMM_WORLD); //Everyone has a list of size INIT_SIZE (minus collisions?) 
  
  inittime = MPI_Wtime()-starttime;
  if(rank == 0) fprintf(rout, "Barrier passed. (%.2lf)\n",inittime);
  
  iterations = 0;
  while(1) { //The main loop
    iterations++;
    //Root samples all vectors
    if(rank == 0){
      j=0;
      while(stacksize > 0 && j < Qsize){
        copy_element((Q+j), (S+stacksize-1));
        stacksize--; 
        copy((com_buffer+j*N), Q[j].v,N);
        j++;        
      }
      for(; j < Qsize; j++){
        sample(Q[j].v, b,mu,vars,t);
        copy((com_buffer+j*N), Q[j].v,N);
      }      
    }
    
    MPI_Bcast (com_buffer,(2*N*Qsize),MPI_BYTE,0,MPI_COMM_WORLD); //Samples are broadcast
    
    if(rank > 0){
      for(i=0;i<Qsize;i++){
        copy(Q[i].v,(com_buffer+i*N),N);
        Q[i].inprod = innerproduct(Q[i].v,Q[i].v);
        for(l=0; l<N; l++) x[N-1-l] = Q[i].v[l];
        toFFT(yy1,x);
        FFT2SSE1(Q[i].fft,yy1);
      }
    } else {
      for(i=0;i<Qsize;i++){
        Q[i].inprod = innerproduct(Q[i].v,Q[i].v);
        for(l=0; l<N; l++) x[N-1-l] = Q[i].v[l];
        toFFT(yy1,x);
        FFT2SSE1(Q[i].fft,yy1);
      }
    }
    
    //Prep survivor mask
    for(l=0; l < Qsize; l++){
      setbit(survivors, l);
    }
    for(l = Qsize; l < 64*(Qsize/64 + 1); l++){
      clearbit(survivors, l);
    }
    
    //Reduce all samples <-> list.
    for(i = 0; i < Lsize; i++){
      j = 0;
      if(getbit(L_bits, i) != 0){
        j++;
        //Take L[i] and <-> it with the samples, no insertions
        for(k = 0; k < Qsize; k++){
          if(getbit(survivors,k) != 0){
            int red = reduce((L+i),(Q+k));

            if (red & 1) { 
              clearbit (L_bits, i); 
              listsize--; j--; 
              if (L[i].inprod == 0) { collisions++; } 
              else {
                //Save on local stack, send to root at end
                copy_element((S+stacksize),(L+i));
                INCREMENT_STACKSIZE
              }
              break;
            } else if (red & 2) { 
              clearbit(survivors, k);
              if (Q[k].inprod == 0) { 
                collisions++;
              }
            }            
          }
        }
      }
      if(j == listsize) break;
    }

//Reduce other samples <-> your samples
#define SAMP2SAMP(k,i) \
if (getbit(survivors, i) != 0) {        \
  int red = reduce((Q+k),(Q+i));        \
  if (red & 1) {                        \
    clearbit(survivors, k);             \
    if (Q[k].inprod == 0) collisions++; \
    break;                              \
  } else if (red & 2) {                 \
    clearbit (survivors, i);            \
    if (Q[i].inprod == 0) collisions++; \
  }                                     \
}
    
    /* Sample at position k is reduced 
     * wrt all samples of same parity before
     * it and all samples of differing parity after it.
     * Your samples must be in forward FFT and listrep.
     */
    for(k = Qoffset; k < QoffsetB; k++){
      if(getbit(survivors,k) != 0){
        //Change FFT of Q[k] to forward and listrep
        toFFT(yy1,Q[k].v);
        FFT2SSE2(Q[k].fft,yy1);
        
        for(i = k%2; i < Qoffset; i += 2){
          SAMP2SAMP(k,i);
        }
        if(getbit(survivors,k) == 0) continue;
        for(i=k+1; i < QoffsetB; i++){
          SAMP2SAMP(k,i);
        }
        if(getbit(survivors,k) == 0) continue;
        for(i=QoffsetB+((QoffsetB-k)%2); i < Qsize; i+=2){
          SAMP2SAMP(k,i);
        }
      }
    }
        
    //Collectively figure out who survived
    MPI_Allreduce (survivors,rsurv,8*(Qsize/64+1),MPI_BYTE,MPI_BAND,MPI_COMM_WORLD); 
    
    lrsend.lsize = listsize;
    lrsend.rank = rank;
    
    //Check whether the computation should be terminated
    if(terminate > 0 || collisions > 0.1*listsize+200 || MPI_Wtime() - starttime > 3600){
      lrsend.lsize = -1;
    }
    
    //Collectively compute who has the smallest list (tiebreaker being smaller rank)
    MPI_Allreduce (&lrsend,&lrrecv,1,MPI_2INT,MPI_MINLOC,MPI_COMM_WORLD); 
    
    if(lrrecv.lsize < 0) break; // At least one node hit the terminate condition
    
    if(rank == lrrecv.rank){ //Insert the survivors into your list
      i=0;
      for(k=0; k < Qsize; k++){
        if(getbit(rsurv,k) != 0){
          for(; i < Lsize; i++){
            if(getbit(L_bits,i) == 0){
              setbit(L_bits,i);
              copy_element((L+i),(Q+k));
              toFFT(yy1,L[i].v);
              FFT2SSE2(L[i].fft,yy1);
              if(L[i].inprod < minimum){
                minimum = L[i].inprod;
                mintime = MPI_Wtime() - starttime;
                printf("Rank %d: New minimum %lu (%.2lf)\n", rank, minimum, mintime); fflush(stdout);
                if(minimum < goal) {
                  if(ghtime < 0) ghtime = MPI_Wtime()-starttime;
                  printf("Goal reached! Entering termination.\n");
                  copy(svp,L[i].v,N);
                  terminate = 1;
                }
              }
              INCREMENT_LISTSIZE
              break;
            }
          }
        }
      }
    }
    
    for(i=0; i < Qsize; i++){
      irsend[i].inprod = Q[i].inprod;
      irsend[i].rank = rank;
    }
    
    //Root computes who has smallest innerproduct for non-survivors.
    MPI_Reduce(irsend,irrecv,Qsize,MPI_LONG_INT,MPI_MINLOC,0,MPI_COMM_WORLD);
    
    if(rank == 0){ //Root prepares mask for which vectors to request from nodes
      for(i=0; i < size; i++){
        reqnums[i] = 0;
        for(j=0;j < (Qsize/64+1); j++) reqsurv[i*(Qsize/64+1)+j] = 0;
      }
      
      for(i=0;i<Qsize;i++){
        if(getbit(rsurv,i) == 0 && irrecv[i].inprod > 0){
          j = irrecv[i].rank;
          reqnums[j]++;
          setbit(reqsurv,j*(64)*(Qsize/64+1)+i);
        }
      }
    }
    //Root sends mask for which vectors to request from each node
    MPI_Scatter (reqsurv,8*(Qsize/64+1),MPI_BYTE,reqrecv,8*(Qsize/64+1),MPI_BYTE,0,MPI_COMM_WORLD);
    
    //Consolidate all non-surviving samples in the root-stack
    if(rank == 0){ //Root receives everyone's 
      for(i = 0; i < Qsize; i++){
        if(getbit(reqrecv,i) != 0){
          copy_element((S+stacksize),(Q+i));
          INCREMENT_STACKSIZE
        }
      }
      for(i = 1; i < size; i++){
        if(reqnums[i] > 0){
          MPI_Recv(com_buffer,2*(N*reqnums[i]),MPI_BYTE,i,MPI_ANY_TAG,MPI_COMM_WORLD,&status);
          for(j=0; j < reqnums[i]; j++){
            copy(S[stacksize].v,(com_buffer+j*N),N);
            S[stacksize].inprod = innerproduct(S[stacksize].v,S[stacksize].v); //TODO: Can this be helped?
            INCREMENT_STACKSIZE
          }
        }
      }
      
    } else {
      j=0;
      for(i=0; i < Qsize; i++){
        if(getbit(reqrecv,i) != 0){
          copy((com_buffer+j*N),Q[i].v,N);
          j++;
        }
      }
      if(j>0) MPI_Send (com_buffer,2*(N*j),MPI_BYTE,0,0,MPI_COMM_WORLD);
    }
    
    //Root gathers everyone's stacksizes
    MPI_Gather (&stacksize,1,MPI_INT,reqnums,1,MPI_INT,0,MPI_COMM_WORLD);
    
    if(rank == 0){ //Root receives everyone's stack contents
      for(i=1; i < size; i++){
        int incstacksize = reqnums[i];
        if(incstacksize > 0){
          if(N*incstacksize > comSize){
            comSize = N*incstacksize;
            com_buffer = (int16_t *) realloc (com_buffer, (comSize) * sizeof (int16_t));
          }
          MPI_Recv(com_buffer,2*(N*incstacksize),MPI_BYTE,i,MPI_ANY_TAG,MPI_COMM_WORLD,&status);
          for(k=0; k < incstacksize; k++){
            copy(S[stacksize].v,(com_buffer+k*N),N);
            S[stacksize].inprod = innerproduct(S[stacksize].v,S[stacksize].v);
            INCREMENT_STACKSIZE
          }
        }
      }
    } else if (stacksize>0) { //Everyone else sends their stack to the root
      if(N*stacksize > comSize){
        comSize = N*stacksize;
        com_buffer = (int16_t *) realloc (com_buffer, (comSize) * sizeof (int16_t));
      }
      for(k=0; k < stacksize; k++){
        copy((com_buffer+k*N),S[k].v,N);
      }
      MPI_Send (com_buffer,(2*N*stacksize),MPI_BYTE,0,0,MPI_COMM_WORLD);
      stacksize = 0;
    }
        
    if(iterations%128 == 0) {
      int tlist;
      MPI_Reduce (&listsize,&tlist,1,MPI_INT,MPI_SUM,0,MPI_COMM_WORLD);     
      if(rank == 0) { 
        fprintf(rout, "%d %d %2.lf\n", iterations, tlist, MPI_Wtime() - starttime); 
        fflush(rout); 
      }
    }

    /* Enable the code below if storing the state is required. */
#if STORE
    if(iterations%512 == 0) {
      //printf ("Rank: %d Stack size %d, Ssize, %d, List size %d, Lsize %d, min %d\n", rank, stacksize, Ssize, listsize, Lsize, minimum);
      //printf("Writing state to disk for rank %d (%.2lf)\n", rank, MPI_Wtime()-starttime);
      store_state (fname, stacksize, Ssize, S, \
                   listsize, Lsize, oldLsize, L, L_bits, minimum, iterations, collisions);
    }
#endif

  }

  {
    int tlist, tstack, tcoll;
    //long tmin;
    double minght;
    if(ghtime < 0 || ghtime == 0.0) ghtime = MPI_Wtime() - starttime;
    MPI_Reduce (&listsize,&tlist,1,MPI_INT,MPI_SUM,0,MPI_COMM_WORLD);
    MPI_Reduce (&stacksize,&tstack,1,MPI_INT,MPI_SUM,0,MPI_COMM_WORLD);
    MPI_Reduce (&collisions,&tcoll,1,MPI_INT,MPI_SUM,0,MPI_COMM_WORLD);
    MPI_Reduce (&ghtime,&minght,1,MPI_DOUBLE,MPI_MIN,0,MPI_COMM_WORLD);
    if(rank == 0) { fprintf(rout, "R = %d, TL = %d, TC = %d, TS = %d (T %2.lf, GHT %2.lf)\n", iterations, tlist, tcoll, tstack, MPI_Wtime() - starttime, minght); fflush(rout); }
    minsend.inprod = minimum;
    minsend.rank = rank;
    MPI_Allreduce(&minsend, &minrecv, 1, MPI_LONG_INT,MPI_MINLOC, MPI_COMM_WORLD);
    if(rank == minrecv.rank) {
      FILE *file;
      file = fopen("Min.out", "w");
      fprintf(file, "%lu (%.2lf)\n", minimum, mintime);
      fprintf(file, "[");
      for(l=0;l<N;l++) fprintf(file, "%d ", svp[l]);
      fprintf(file,"]\n"); fflush(file);
    }
  }
  FFT_free();
  
  return 1;
}

int main (int argc, char* argv[]) {
  int rank, size;

  if (argc != 2) {
    printf ("Usage %s filename\n", argv[0]);
    return -1;
  }
  /* We only support dim. 128. */
  N = 128;
  n = N;
  
  fileName = (char *) malloc (64 * sizeof(char));
  fileName = argv[1];
  
  MPI_Init(&argc, &argv);
  MPI_Comm_rank(MPI_COMM_WORLD,&rank);
  MPI_Comm_size(MPI_COMM_WORLD,&size);
  
  if(rank == 0) {
    rout = fopen("Root.out", "a");
  }
  client (rank,size);

  MPI_Barrier(MPI_COMM_WORLD);
  MPI_Finalize();
  return 0;
}

