/* This is free and unencumbered software released into the public domain.
 * 
 * Anyone is free to copy, modify, publish, use, compile, sell, or
 * distribute this software, either in source code form or as a compiled
 * binary, for any purpose, commercial or non-commercial, and by any
 * means.
 *
 * See UNLICENSE for complete information.
 *
 * This code implements our modified Gauss-sieve algorithm specifically 
 * aimed at finding short vectors in an ideal lattice. Ideal lattices are 
 * lattices with additional structure, namely they are also ideals in a ring R.
 * We restrict to lattices which correspond to ideals in the ring of integers 
 * of a cyclotomic number field, namely R = Z[X]/(X^(2^n)+1). We use this structure
 * by exploting the fact that rotations of the element of the vector are in the
 * lattice as well. This implementation works for dimensions 2^n >= 8. 
 *
 * See for a detailed description of the methods used:
 * Joppe W. Bos, Michael Naehrig, and Joop van de Pol:
 * Sieving for Shortest Vectors in Ideal Lattices:
 * a Practical Perspective. Cryptology ePrint Archive, 
 * Report 2014/880, IACR, 2014.
 * http://eprint.iacr.org/2014/880.pdf

 */

#include "shared.h"

/* Saves the list every STORE rounds or never if 0 */
#define STORE 0

/* Size increments of main list and stack respectively */
#define SIZE (8*1024)
#define SIZE2 1024

/* Starting list size */
#define INIT_SIZE (128)

/* Samples to be used per round per node */
#define BATCH (16)

FILE *rout;

typedef struct {
  __m128i *v;
  IP inprod;
} element_t;

element_t *tmp;

static void element_init (element_t *new) {
  posix_memalign ((void *) &(new->v), 16, (N/8) * sizeof (__m128i));
  if (new->v == NULL) {
    fprintf (stderr, "Error: element_init out of memory.\n");
    exit (-1);
  }
}

static void copy_element (element_t *c, element_t *a) {
  memcpy (c->v, a->v, (N/8)*sizeof(__m128i));
  c->inprod = a->inprod;
}

static IP innerproduct (__m128i *a, __m128i *b) {
  __m128i c, sum;
  static int64_t t[2] __attribute__ ((aligned (16)));
  int i;
  static int init = 1;
  static __m128i zero;
  __m128i cmp;

  if (init) {
    uint32_t tmp[4];

    tmp[0] = (uint32_t) 0;
    tmp[1] = (uint32_t) 0;
    tmp[2] = (uint32_t) 0;
    tmp[3] = (uint32_t) 0;
    zero = _mm_loadu_si128 ((__m128i *) tmp);

    init = 0;
  }

  c = _mm_madd_epi16 (a[0], b[0]);
  cmp = _mm_cmplt_epi32 (c, zero);
  sum = _mm_add_epi64 (_mm_unpacklo_epi32 (c, cmp), _mm_unpackhi_epi32 (c, cmp));

  for (i=1; i < N/8; i++) {
    c = _mm_madd_epi16 (a[i], b[i]);
    cmp = _mm_cmplt_epi32 (c, zero);
    sum = _mm_add_epi64 (sum, _mm_add_epi64 (_mm_unpacklo_epi32 (c, cmp), _mm_unpackhi_epi32 (c, cmp)));
  }
  _mm_store_si128 ((__m128i *) t, sum);

  return (t[0]+t[1]);
}

static void rotate(__m128i *b, __m128i *a){
  int i = N/8 - 1;
  __m128i c;
  c = _mm_slli_si128 (a[0],14);

  b[i] = _mm_subs_epi16 (_mm_srli_si128(a[i],2),c);

  c = _mm_slli_si128 (a[i],14);
  for(--i; i >= 0; i--){
    b[i] = _mm_add_epi16 (_mm_srli_si128(a[i],2),c);
    c = _mm_slli_si128 (a[i],14);
  }
}

/* b is rotated */
static IP innerproduct_rotate_right (__m128i *a, __m128i *b, int index) {
  __m128i c, sum;
  static int64_t t[2] __attribute__ ((aligned (16)));
  int i;
  static int init = 1;
  static __m128i zero;
  __m128i cmp;

  if (init) {
    uint32_t tmp[4];

    tmp[0] = (uint32_t) 0;
    tmp[1] = (uint32_t) 0;
    tmp[2] = (uint32_t) 0;
    tmp[3] = (uint32_t) 0;
    zero = _mm_loadu_si128 ((__m128i *) tmp);

    init = 0;
  }

  c = _mm_madd_epi16 (a[N/8-1], b[N/8-1-index]);
  cmp = _mm_cmplt_epi32 (c, zero);
  sum = _mm_add_epi64 (_mm_unpacklo_epi32 (c, cmp), _mm_unpackhi_epi32 (c, cmp));

  for (i=(N/8)-2; i >= index; i--) {
    c = _mm_madd_epi16 (a[i], b[i-index]);
    cmp = _mm_cmplt_epi32 (c, zero);
    sum = _mm_add_epi64 (sum, _mm_add_epi64 (_mm_unpacklo_epi32 (c, cmp), _mm_unpackhi_epi32 (c, cmp)));
  }
  for (; i >= 0; i--) {
    c = _mm_madd_epi16 (a[i], _mm_sub_epi16 (zero, b[i-index+N/8]));
    cmp = _mm_cmplt_epi32 (c, zero);
    sum = _mm_add_epi64 (sum, _mm_add_epi64 (_mm_unpacklo_epi32 (c, cmp), _mm_unpackhi_epi32 (c, cmp)));
  }

  _mm_store_si128 ((__m128i *) t, sum);

  return (t[0]+t[1]);
}

/* a is rotated */
static void vec_submul (__m128i *res, INT c, __m128i *a, int index) {
  int i, j;
  
  if (c > 0) {
    for (i=0; i < c; i++) {
      for (j=0; j < N/8 - index; j++) {  
        res[j] = _mm_sub_epi16 (res[j], a[j+index]);
      }
      for(;j<N/8;j++){
        res[j] = _mm_add_epi16 (res[j], a[j+index-N/8]);
      }
    }
  } else {
    for (i=0; i < -c; i++) {
      for (j=0; j < N/8 - index; j++) { 
        res[j] = _mm_add_epi16 (res[j], a[j+index]);
      }
      for(;j<N/8;j++){
        res[j] = _mm_sub_epi16 (res[j], a[j+index-N/8]);
      }
    }
  }
}

static void vec_submul_right (__m128i *res, INT c, __m128i *a, int index) {
  int i, j;

  if (c > 0) {
    for (i=0; i < c; i++) {
      for (j=0; j < index; j++) {  
        res[j] = _mm_add_epi16 (res[j], a[j-index+N/8]);
      }
      for(;j<N/8;j++){
        res[j] = _mm_sub_epi16 (res[j], a[j-index]);
      }
    }
  } else {
    for (i=0; i < -c; i++) {
      for (j=0; j < index; j++) { 
        res[j] = _mm_sub_epi16 (res[j], a[j-index+N/8]);
      }
      for(;j<N/8;j++){
        res[j] = _mm_add_epi16 (res[j], a[j-index]);
      }
    }
  }
}

static int reduce (element_t *list, element_t *samp, __m128i **rotations) {
  IP inprodnew, inprodnewa; 
  int I, J, order; 
  
  order = (samp->inprod < list->inprod);
  for (I=0; I < 8; I++) { 
    for (J=0; J < N/8; J++) { 
      inprodnew = innerproduct_rotate_right (rotations[I], samp->v, J);
      inprodnewa = (inprodnew < 0) ? -inprodnew : inprodnew; 
      inprodnewa <<= 1; 
      if (order && inprodnewa > samp->inprod) {
        INT mu; 
        mu = ((inprodnew^samp->inprod)<0); mu = -mu*2+1; 
        vec_submul_right (rotations[I], mu, samp->v, J); //FIX
        UPDATE_MU_STAR(list,samp) 
        memcpy(list->v, rotations[I], (N/8)*sizeof(__m128i));
        //if(list->inprod != innerproduct(list->v,list->v)) printf("Error!\n");
        
        //Vector list has been reduced
        return 1;
      } else if (!order && inprodnewa > list->inprod) { 
        INT mu; 
        mu = ((inprodnew^samp->inprod)<0); mu = -mu*2+1; 
        vec_submul (samp->v, mu, rotations[I], J); 
        UPDATE_MU_STAR(samp,list)
        //if(samp->inprod != innerproduct(samp->v,samp->v)) printf("Error2!\n");
        
        //Vector samp has been reduced
        return 2;
      }
    } 
  }
  return 0;
}

#if STORE
static void store_element (element_t *element, FILE *out) {
  fwrite (element->v, sizeof (__m128i), (N/8), out);
  fwrite (&element->inprod, sizeof(IP), 1, out);
}

static int store_state (char *fname, int stacksize, int Ssize, element_t *S, \
                 int listsize, int Lsize, int oldLsize, element_t *L, \
                 int64_t *L_bits, uint64_t minimum, int iterations, int collisions) {
  char str[1024];
  FILE *out;
  int i;

  /* Keep always an old and a new copy around just to be sure. */
  sprintf (str, "(mv %s %s.old 2>&1) >> /dev/null", fname, fname);
  system (str);

  out = fopen (fname, "w");
  if (out == NULL) return -1;

  /* Write the stack. */
  fwrite (&stacksize, sizeof(int), 1, out);
  fwrite (&Ssize, sizeof(int), 1, out);
  for (i=0; i < stacksize; i++) {
    store_element (&S[i], out);
  }

  /* Write the list. */
  fwrite (&listsize, sizeof(int), 1, out);
  fwrite (&Lsize, sizeof(int), 1, out);
  fwrite (&oldLsize, sizeof(int), 1, out);
  for (i=0; i < listsize; i++) {
    store_element (&L[i], out);
    //printf ("Store: %ld\n", L[i].inprod);
  }

  fwrite (L_bits, sizeof(int64_t), (Lsize/64+1), out);

  fwrite (&minimum, sizeof(uint64_t), 1, out);
  fwrite (&iterations, sizeof(int), 1, out);
  fwrite (&collisions, sizeof(int), 1, out);

  fclose (out);  
  return 1;
}
#endif

static void load_element (element_t *element, FILE *out) {
  fread (element->v, sizeof (__m128i), (N/8), out);
  fread (&element->inprod, sizeof(IP), 1, out);
  if (innerproduct(element->v,element->v) != element->inprod) {
    printf ("Read error\n");
    exit (0);
  }
}

static int load_state (char *fname, int *stacksize, int *Ssize, element_t **S, \
                int *listsize, int *Lsize, int *oldLsize, element_t **L, \
                int64_t **L_bits, int64_t *minimum, int *iterations, int *collisions) {
  FILE *out;
  int i;

  out = fopen (fname, "r");
  if (out == NULL) return -1;

  /* Read the stack. */
  fread (stacksize, sizeof(int), 1, out);
  fread (Ssize, sizeof(int), 1, out);
  (*S) = (element_t *) malloc ((*Ssize) * sizeof (element_t));
  for(i=0; i<(*Ssize); i++) element_init (&(*S)[i]);
  for (i=0; i<(*stacksize); i++) {
    load_element (&(*S)[i], out);
  }

  /* Read the list. */
  fread (listsize, sizeof(int), 1, out);
  fread (Lsize, sizeof(int), 1, out);
  fread (oldLsize, sizeof(int), 1, out);
  (*L) = (element_t *) malloc ((*Lsize) * sizeof (element_t));
  for(i=0; i<(*Lsize); i++) element_init (&(*L)[i]);
  for (i=0; i<(*listsize); i++) {
    load_element (&(*L)[i], out);
  }

  *L_bits = (int64_t *) malloc (((*Lsize)/64+1)*sizeof(int64_t));
  fread (*L_bits, sizeof(int64_t), ((*Lsize)/64+1), out);

  fread (minimum, sizeof(uint64_t), 1, out);
  fread (iterations, sizeof(int), 1, out);
  fread (collisions, sizeof(int), 1, out);

  fclose (out);  
  return 1;
}

int client(int rank, int size){
  MPI_Status status;
  INT **b; //Basis, used for sampling
  INT *sampv; //Vector, used for sampling
  element_t *L = NULL; //List
  int64_t *L_bits = NULL; //Mask indicating list occupancy
  int listsize; //Number of vectors currently in the list
  int Lsize, oldLsize; //Current list storage size
  
  element_t *S = NULL; //Stack
  int stacksize; //Number of vectors currently in the stack
  int Ssize; //Current stack storage size
  element_t new;
  
  element_t *Q = NULL; //Sample list
  int Qsize = BATCH*size;
  int Qoffset = BATCH*rank;
  int QoffsetB = Qoffset+BATCH;
  
  __m128i *com_buffer = NULL; //Used to send vectors
  int comSize;
  int64_t *survivors = NULL, *rsurv = NULL; //Masks indicating survivors
  int64_t *reqsurv = NULL, *reqrecv = NULL; //Masks for requesting vectors
  int *reqnums = NULL;
  
  int16_t svp[N];
  int i, j, k, l;
  int64_t max, minimum, goal;
  double **gs, **mu, *lbj = NULL, *vars = NULL, t; //Gram-Schmidt used for sampling
  int collisions, iterations;
  int terminate;
  double starttime, inittime, mintime;
  __m128i **rotations;
  
  struct { 
    int lsize; 
    int rank; 
  } lrsend, lrrecv;

  struct _inprod_rank {
    int64_t inprod;
    int32_t rank;
  } *irsend = NULL, *irrecv = NULL, minsend, minrecv;
 
  char fname[1024];
  int restore = 0;

  irsend = (struct _inprod_rank*) malloc (Qsize * sizeof(struct _inprod_rank));
  irrecv = (struct _inprod_rank*) malloc (Qsize * sizeof(struct _inprod_rank));
  reqnums = (int32_t *) malloc (size * sizeof(int32_t));
  if (irsend == NULL || irrecv == NULL || reqnums == NULL) {
    fprintf (stderr, "Memory error\n");
    return -1;
  }

  sprintf (fname, "PN_store.rank.%d", rank);

  //Randomize seed
  {
    struct timeval time; 
    gettimeofday(&time,NULL);
    srand( ((rank << 16) * getpid()) ^ ((time.tv_sec * 1000) + (time.tv_usec / 1000)) ); 
  }

  element_init(&new);

  /* Try to open a stored state. */
  {
    FILE *try = fopen (fname, "r");
    if (try != NULL) {
      restore = 1;
      fclose (try);
    }
  }
  
  if (rank == 0) { fprintf(rout,"Start minimum %ld (goal %ld)\n", minimum, goal); fflush(rout); }

  if (restore) {
    load_state (fname, &stacksize, &Ssize, &S, \
                &listsize, &Lsize, &oldLsize, &L, &L_bits, &minimum, &iterations, &collisions);
  } else {
    //Initialize memory
    Lsize = SIZE;
    oldLsize = 0;
    L = (element_t *) malloc (SIZE * sizeof (element_t));
    for(i=0; i < SIZE; i++){
      element_init((L+i));
    }

    L_bits = (int64_t *) calloc ((SIZE/64+1), sizeof(int64_t));
  
    Ssize = SIZE2;
    S = (element_t *) malloc (SIZE2 * sizeof (element_t));
    for(i=0; i < SIZE2; i++){
      element_init((S+i));
    }
  }

#define INCREMENT_LISTSIZE                                             \
listsize++;                                                            \
if(listsize == Lsize) {                                                \
  oldLsize = Lsize;                                                    \
  Lsize += SIZE;                                                       \
  L = (element_t *) realloc (L, Lsize*sizeof(element_t));              \
  for(k=oldLsize; k < Lsize; k++) element_init(L+k);                   \
  L_bits = (int64_t *) realloc (L_bits, (Lsize/64+1)*sizeof(int64_t)); \
  for (l=oldLsize; l < Lsize; l++) { clearbit(L_bits,l); }             \
}

#define INCREMENT_STACKSIZE                                    \
stacksize++;                                                   \
if(stacksize == Ssize) {                                       \
  l = Ssize;                                                   \
  Ssize += SIZE2;                                              \
  S = (element_t *) realloc (S, (Ssize) * sizeof (element_t)); \
  for(;l<Ssize; l++) element_init((S+l));                      \
}

  Q = (element_t *) malloc (Qsize * sizeof (element_t));
  for(i=0; i < Qsize; i++){
    element_init((Q+i));
  }
  Qoffset = BATCH*rank;
  
  comSize = (N/8)*SIZE2;
  posix_memalign ((void *) &com_buffer, 16, (comSize) * sizeof (__m128i)); 
  posix_memalign ((void *) &rotations, 16, 8* sizeof (__m128i));
  for(i=0; i < 8; i++) posix_memalign ((void *) &rotations[i], 16, (N/8)* sizeof (__m128i));
  
  survivors = (int64_t *) malloc ((Qsize/64+1)*sizeof(int64_t));
  rsurv = (int64_t *) malloc ((Qsize/64+1)*sizeof(int64_t));
  reqsurv = (int64_t *) malloc((Qsize/64+1) * size * sizeof(int64_t));
  reqrecv = (int64_t *) malloc((Qsize/64+1) * sizeof(int64_t));
  
  tmp = (element_t *) malloc (sizeof(element_t));
  element_init(tmp);
  
  for(i = 0; i < Qsize; i++) irsend[i].rank = rank;
  
  //Initialize memory for basis
  b = (INT**) malloc (N * sizeof (INT*));
  for (i=0; i < N; i++) {
    b[i]  = (INT*) malloc (N * sizeof (INT));
  }
  posix_memalign ((void *) &(sampv), 16, (N) * sizeof (INT));
  
  max = readbasis (fileName, b);
  if (max == 0) {
    printf ("Read basis error\n");
    exit (0);
    return -1;
  }
  
  gs  = (double **) malloc (N * sizeof (double*));
  mu  = (double **) malloc (N * sizeof (double*));
  lbj = (double *)  malloc (N * sizeof (double));
  vars = (double *)  malloc (N * sizeof (double));
        
  for (i=0; i < N; i++) {
    gs[i] = (double *)  malloc (N * sizeof (double));
    mu[i] = (double *)  malloc (N * sizeof (double));
  }

  gram_schmidt (gs, mu, lbj, b, &goal);
  sample_init(lbj,vars);

  t=(log((double)N)*log((double)N))/70.; // ToDO: find the optimal value for t  
  
  terminate = 0;  
  starttime = MPI_Wtime();
  mintime = 0;

  if (!restore) {
    toSSE2(new.v, b[0]);
    minimum = innerproduct(new.v,new.v);
    //Populate the list with the first element 
    if(rank < N){
      toSSE2(L[0].v,b[rank]);
    } else {
      sample (sampv,b,mu,vars,t);
      toSSE2(L[0].v,sampv);
    }
    L[0].inprod = innerproduct(L[0].v,L[0].v); 
    listsize = 1;
  
    if (rank == 0) { fprintf(rout,"Start minimum %lu\n", minimum); fflush(rout); }
    //Now fill the stack up to INIT_SIZE-1
    stacksize = 0;
    for(i=rank+size; i < N; i+=size){
      toSSE2(S[stacksize].v,b[i]);
      S[stacksize].inprod = innerproduct(S[stacksize].v,S[stacksize].v);
      stacksize++;
    }
  
    for(;stacksize < INIT_SIZE-1; stacksize++){
      sample (sampv,b,mu,vars,t);
      toSSE2(S[stacksize].v,sampv);
      S[stacksize].inprod = innerproduct(S[stacksize].v,S[stacksize].v);
    }
    k=0;
  
    //Perform reduction until stack is empty
    collisions = 0;
    while(stacksize > 0){ 
      int firstempty = Lsize;
      j=0;
      copy_element(&new, (S+stacksize-1));
      stacksize--;
      memcpy(rotations[0], new.v, (N/8)*sizeof(__m128i));
      for(l=1;l<8;l++) rotate(rotations[l],rotations[l-1]);
    
      for(i = 0; i < Lsize; i++){
        if(getbit(L_bits, i) != 0){
          int red = reduce(&new,(L+i),rotations);
          j++;
        
          if(red & 1){
            //new has been reduced
            if (new.inprod != 0) { 
              memcpy(rotations[0], new.v, (N/8)*sizeof(__m128i));
              for(l=1;l<8;l++) rotate(rotations[l],rotations[l-1]);
              i=-1; j=0; continue;
            } else { collisions++; break; } 
          } else if (red & 2){
            clearbit(L_bits, i);
            listsize--; j--; 
            if (L[i].inprod == 0) { collisions++; } 
            else { 
              copy_element((S+stacksize),(L+i));
              INCREMENT_STACKSIZE
            }
          }      
        } else if(i < firstempty) firstempty = i;  
        if(j == listsize && (firstempty < Lsize)) break;
      }
      if(new.inprod == 0){
        continue;
      } else if(new.inprod < minimum){
        minimum = new.inprod;
        printf("Rank %d: New minimum %lu\n", rank, minimum);
        if(minimum < goal) {
          printf("Goal reached! Entering termination.\n");
          terminate = 1;
        }
      }
      if (firstempty == Lsize){
      //Error
      fprintf (stderr, "Error: list error.\n");
      exit(1);
      }
      copy_element((L+firstempty), &new);
      setbit(L_bits,firstempty);
      INCREMENT_LISTSIZE
    }
  }
  
  MPI_Barrier (MPI_COMM_WORLD); //Everyone has a list of size INIT_SIZE (minus collisions?)
  
  inittime = MPI_Wtime() - starttime;
  if(rank == 0) { fprintf(rout,"Barrier passed. (%.2lf)\n", inittime); fflush(rout); }
  
  iterations = 0;
  while (1) { //The main loop
    iterations++;
    
    //Root samples all vectors
    if(rank == 0){
      j=0;
      while(stacksize > 0 && j < Qsize){
        copy_element((Q+j), (S+stacksize-1));
        stacksize--; 
        memcpy ((com_buffer+j*(N/8)), Q[j].v, (N/8)*sizeof(__m128i));
        j++;
      }
      for(; j < Qsize; j++){
        sample(sampv, b,mu,vars,t);
        toSSE2(Q[j].v,sampv);
        Q[j].inprod = innerproduct(Q[j].v,Q[j].v);
        memcpy ((com_buffer+j*(N/8)), Q[j].v, (N/8)*sizeof(__m128i));
      }
    }
    
    MPI_Bcast (com_buffer,(2*N*Qsize),MPI_BYTE,0,MPI_COMM_WORLD); //Samples are broadcast
    
    if(rank > 0){
      for(i=0;i<Qsize;i++){
        memcpy (Q[i].v, com_buffer+i*(N/8), (N/8)*sizeof(__m128i));
        Q[i].inprod = innerproduct(Q[i].v,Q[i].v);
      }
    }
    
    //Prep survivor mask
    for(l=0; l < Qsize; l++){
      setbit(survivors, l);
    }
    for(l = Qsize; l < 64*(Qsize/64 + 1); l++){
      clearbit(survivors, l);
    }
    
    //Reduce all samples <-> list.
    for(i = 0; i < Lsize; i++){
      j = 0;
      if(getbit(L_bits, i) != 0){
        j++;
        //Take L[i] and <-> it with the samples, no insertions
      
        memcpy(rotations[0], L[i].v, (N/8)*sizeof(__m128i));
        for(l=1;l<8;l++) rotate(rotations[l],rotations[l-1]);
    
        for(k = 0; k < Qsize; k++){
          if(getbit(survivors,k) != 0){
            int red = reduce((L+i),(Q+k),rotations);
            if (red == 1) {
              clearbit (L_bits, i); 
              if (L[i].inprod == 0) { collisions++; } 
              else {
                //Save on local stack, send to root at end
                copy_element((S+stacksize),(L+i));
                INCREMENT_STACKSIZE
              }
              listsize--; j--; 
              break;
            } else if (red == 2) {
              clearbit(survivors, k);
              if (Q[k].inprod == 0) collisions++;
            }
          }
        }
      }
      if(j == listsize) break;
    }

//Reduce samples <-> your sample
#define SAMP2SAMP(k,i)                     \
if(getbit(survivors, i) != 0) {            \
  int red = reduce((Q+k),(Q+i),rotations); \
  if (red & 1) {                           \
    clearbit(survivors, k);                \
    if (Q[k].inprod == 0) collisions++;    \
    break;                                 \
  } else if (red & 2) {                    \
    clearbit (survivors, i);               \
    if (Q[i].inprod == 0) collisions++;    \
  }                                        \
}
    /* Sample at position k is reduced 
     * wrt all samples of same parity before 
     * it and all samples of differing parity after it
     */
    
    for(k = Qoffset; k < QoffsetB; k++){
      if(getbit(survivors,k) != 0){
        memcpy(rotations[0], Q[k].v, (N/8)*sizeof(__m128i));
        for(l=1;l<8;l++) rotate(rotations[l],rotations[l-1]);
        for(i = k%2; i < Qoffset; i += 2){
          SAMP2SAMP(k,i);
        }
        if(getbit(survivors,k) == 0) continue;
        for(i=k+1; i < QoffsetB; i++){
          SAMP2SAMP(k,i);
        }
        if(getbit(survivors,k) == 0) continue;
        for(i=QoffsetB+((QoffsetB-k)%2); i < Qsize; i+=2){
          SAMP2SAMP(k,i);
        }
      }
    }
#undef SAMP2SAMP
    //Collectively figure out who survived
    MPI_Allreduce (survivors,rsurv,8*(Qsize/64+1),MPI_BYTE,MPI_BAND,MPI_COMM_WORLD); 
    
    lrsend.lsize = listsize;
    lrsend.rank = rank;
    
    if(collisions > 0.1 * listsize + 200) {
      printf("Rank %d: Terminating due to too many collisions!\n",rank);
      terminate = 1;
    }
    
    if(terminate > 0 || MPI_Wtime() - starttime > 3600){
      lrsend.lsize = -1;
    }
    
    //Collectively compute who has the smallest list (tiebreaker being smaller rank)
    MPI_Allreduce (&lrsend,&lrrecv,1,MPI_2INT,MPI_MINLOC,MPI_COMM_WORLD); 
    
    if(lrrecv.lsize < 0) break;
    
    if(rank == lrrecv.rank){ //Insert the survivors into your list
      i=0;
      for(k=0; k < Qsize; k++){
        if(getbit(rsurv,k) != 0){
          for(; i < Lsize; i++){
            if(getbit(L_bits,i) == 0){
              setbit(L_bits,i);
              copy_element((L+i),(Q+k));
              if(L[i].inprod < minimum){
                minimum = L[i].inprod;
                mintime = MPI_Wtime() - starttime;
                printf("Rank %d: New minimum %lu (%.2lf)\n", rank, minimum, mintime); fflush(stdout);
                fromSSE2(svp,L[i].v);
                if(minimum < goal) {
                  printf("Goal reached! Entering termination.\n");
                  printf("[");
                  for(l=0;l<N;l++) printf("%d ", svp[l]);
                  printf("]\n"); fflush(stdout);
                  terminate = 1;
                }
              }
              INCREMENT_LISTSIZE
              break;
            }
          }
        }
      }
    }
    
    for(i=0; i < Qsize; i++){
      irsend[i].inprod = Q[i].inprod;
      irsend[i].rank = rank;
    }
    
    //Root computes who has smallest innerproduct for non-survivors.
    MPI_Reduce(irsend,irrecv,Qsize,MPI_LONG_INT,MPI_MINLOC,0,MPI_COMM_WORLD);
    
    if(rank == 0){ //Root prepares mask for which vectors to request from nodes
      for(i=0; i < size; i++){
        reqnums[i] = 0;
        for(j=0;j < (Qsize/64+1); j++) reqsurv[i*(Qsize/64+1)+j] = 0;
      }
      
      for(i=0;i<Qsize;i++){
        if(getbit(rsurv,i) == 0){
          if(irrecv[i].inprod > 0){
            j = irrecv[i].rank;
            reqnums[j]++;
            setbit(reqsurv,j*(64)*(Qsize/64+1)+i);
          }
        }
      }
    }
    //Root sends mask for which vectors to request from each node
    MPI_Scatter (reqsurv,8*(Qsize/64+1),MPI_BYTE,reqrecv,8*(Qsize/64+1),MPI_BYTE,0,MPI_COMM_WORLD);
    
    //Consolidate all non-surviving samples in the root-stack
    if(rank == 0){ //Root receives everyone's 
      for(i = 0; i < Qsize; i++){
        if(getbit(reqrecv,i) != 0){
          copy_element((S+stacksize),(Q+i));
          INCREMENT_STACKSIZE
        }
      }
      for(i = 1; i < size; i++){
        if(reqnums[i] > 0){
          MPI_Recv(com_buffer,2*(N*reqnums[i]),MPI_BYTE,i,MPI_ANY_TAG,MPI_COMM_WORLD,&status);
          for(j=0; j < reqnums[i]; j++){
            memcpy (S[stacksize].v, com_buffer+j*(N/8), (N/8)*sizeof(__m128i)); 
            S[stacksize].inprod = innerproduct(S[stacksize].v,S[stacksize].v); //TODO: Can this be helped?
            INCREMENT_STACKSIZE
          }
        }
      }
      
    } else {
      j=0;
      for(i=0; i < Qsize; i++){
        if(getbit(reqrecv,i) != 0){
          memcpy (com_buffer+j*(N/8), Q[i].v, (N/8)*sizeof(__m128i));
          j++;
        }
      }
      if(j>0) MPI_Send (com_buffer,2*(N*j),MPI_BYTE,0,0,MPI_COMM_WORLD);
    }
    
    //Root gathers everyone's stacksizes
    MPI_Gather (&stacksize,1,MPI_INT,reqnums,1,MPI_INT,0,MPI_COMM_WORLD);
    
    if(rank == 0){ //Root receives everyone's stack contents
      for(i=1; i < size; i++){
        int incstacksize = reqnums[i];
        if(incstacksize > 0){
          if((N/8)*incstacksize > comSize){
            __m128i *tmp; 
            int newcomSize = (N/8)*incstacksize;            
            posix_memalign ((void *) &tmp, 16, (newcomSize) * sizeof (__m128i));
            free(com_buffer);
            com_buffer = tmp;
            comSize = newcomSize;
          }
          MPI_Recv(com_buffer,2*(N*incstacksize),MPI_BYTE,i,MPI_ANY_TAG,MPI_COMM_WORLD,&status);
          for(k=0; k < incstacksize; k++){
            memcpy (S[stacksize].v, com_buffer+k*(N/8), (N/8)*sizeof(__m128i));
            S[stacksize].inprod = innerproduct(S[stacksize].v,S[stacksize].v);
            INCREMENT_STACKSIZE
          }
        }
      }
    } else if (stacksize>0) { //Everyone else sends their stack to the root
      if((N/8)*stacksize > comSize){
        __m128i *tmp; 
        int newcomSize = (N/8)*stacksize;
        posix_memalign ((void *) &tmp, 16, (newcomSize) * sizeof (__m128i));
        free(com_buffer);
        com_buffer = tmp;
        comSize = newcomSize;
      }
      for(k=0; k < stacksize; k++){
        memcpy (com_buffer+k*(N/8), S[k].v, (N/8)*sizeof(__m128i));
      }
      MPI_Send (com_buffer,(2*N*stacksize),MPI_BYTE,0,0,MPI_COMM_WORLD);
      stacksize = 0;
    }
    
    if(iterations%128 == 0) {
      int tlist;
      MPI_Reduce (&listsize,&tlist,1,MPI_INT,MPI_SUM,0,MPI_COMM_WORLD);     
      if(rank == 0) { fprintf(rout, "%d %d %2.lf\n", iterations, tlist, MPI_Wtime() - starttime); fflush(rout); }
    }  


/* Enable the code below if storing the state is required. */
#if STORE
    if(iterations%256 == 0) {
      //printf ("Rank: %d Stack size %d, Ssize, %d, List size %d, Lsize %d, min %d\n", rank, stacksize, Ssize, listsize, Lsize, minimum);
      //printf("Writing state to disk for rank %d (%.2lf)\n", rank, MPI_Wtime()-starttime);
      store_state (fname, stacksize, Ssize, S, \
                   listsize, Lsize, oldLsize, L, L_bits, minimum, iterations, collisions);
    }  
#endif 
  }

  {
    int tlist, tstack, tcoll;
    MPI_Reduce (&listsize,&tlist,1,MPI_INT,MPI_SUM,0,MPI_COMM_WORLD);
    MPI_Reduce (&stacksize,&tstack,1,MPI_INT,MPI_SUM,0,MPI_COMM_WORLD);
    MPI_Reduce (&collisions,&tcoll,1,MPI_INT,MPI_SUM,0,MPI_COMM_WORLD);
    if(rank == 0) { fprintf(rout, "R = %d, TL = %d, TC = %d, TS = %d (T %2.lf)\n", iterations, tlist, tcoll, tstack, MPI_Wtime() - starttime); fflush(rout); }
    minsend.inprod = minimum;
    minsend.rank = rank;
    MPI_Allreduce(&minsend, &minrecv, 1, MPI_LONG_INT,MPI_MINLOC, MPI_COMM_WORLD);
    if(rank == minrecv.rank) {
      FILE *file;
      file = fopen("Min.out", "w");
      fprintf(file, "%lu (%.2lf)\n", minimum, mintime);
      fprintf(file, "[");
      for(l=0;l<N;l++) fprintf(file, "%d ", svp[l]);
      fprintf(file,"]\n"); fflush(file);
    }
  }
    
  return 1;
}

int main (int argc, char* argv[]) {
  int rank, size;
  int ok = 1;

  if (argc != 3) {
    printf ("Usage %s dim filename\nWhere dim is a power of two that is at least 8.\n", argv[0]);
    return -1;
  }
  N = atoi (argv[1]);
  n = N;
  while (n != 1) {
    if (n%2 == 1) { ok = 0; break; }
    n /= 2;
  }  
  if (!ok || N < 8) {
    printf ("Usage %s dim\nWhere dim is a power of two that is at least 8.\n", argv[0]);
    return -1;
  }
  n = N;
  fileName = (char *) malloc (64 * sizeof(char));
  fileName = argv[2];

  MPI_Init(&argc, &argv);
  MPI_Comm_rank(MPI_COMM_WORLD,&rank);
  MPI_Comm_size(MPI_COMM_WORLD,&size);
   
  if(rank == 0) {
    rout = fopen("Root.out", "a");
  }
  client (rank,size);

  MPI_Barrier(MPI_COMM_WORLD);
  MPI_Finalize();
  return 0;
}

